/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <vconf.h>
#include <sys/types.h>
#include <libsyscommon/dbus-system.h>

#include "ambient-mode.h"
#include "util.h"
#include "core.h"
#include "poll.h"
#include "device-node.h"
#include "display-actor.h"
#include "display-ops.h"
#include "core/common.h"
#include "core/devices.h"
#include "core/device-notifier.h"
#include "power/power-handler.h"
#include "led/touch-key.h"

#include <linux/input.h>
#ifndef KEY_SCREENLOCK
#define KEY_SCREENLOCK		0x98
#endif
#ifndef SW_GLOVE
#define SW_GLOVE		0x16
#endif

#define PREDEF_LEAVESLEEP	"leavesleep"
#define POWEROFF_ACT			"poweroff"
#define PWROFF_POPUP_ACT		"pwroff-popup"
#define USEC_PER_SEC			1000000

#define CAPTURE_COMBINATION_INTERVAL		0.5	/* 0.5 second */

#define KEY_MAX_DELAY_TIME		700	/* ms */

#define KEY_RELEASED		0
#define KEY_PRESSED		1
#define KEY_BEING_PRESSED	2

#define SIGNAL_CHANGE_HARDKEY		"ChangeHardkey"
#define SIGNAL_LCDON_BY_POWERKEY	"LCDOnByPowerkey"
#define SIGNAL_LCDOFF_BY_POWERKEY	"LCDOffByPowerkey"

#define TOUCH_RELEASE		(-1)

#define GLOVE_MODE	1

enum key_combination_flags {
	KEY_COMBINATION_STOP		= 0,
	KEY_COMBINATION_POWERKEY	= BIT(0),
	KEY_COMBINATION_MENUKEY		= BIT(1),
};

enum combination_process {
	COMBINATION_STOP	= KEY_COMBINATION_STOP,
	COMBINATION_SCREENCAPTURE	= KEY_COMBINATION_POWERKEY | KEY_COMBINATION_MENUKEY,
};

int __WEAK__ get_glove_state(void);
void __WEAK__ switch_glove_key(int val);

static struct timeval pressed_time;
static guint longkey_timeout_id = 0;
static int cancel_lcdoff;
static int key_combination = KEY_COMBINATION_STOP;
static double combination_pressed_time;
static bool touch_pressed = false;
static int skip_lcd_off = false;
static int skip_combination = false;
static const struct device_ops *touchled;

static int booting_done(void *data)
{
	static int done = 0;

	if (!data)
		return done;

	done = *(int *)data;

	_I("Booting done(%d)", done);

	return done;
}

static inline int current_state_in_on(void)
{
	return (pm_cur_state == S_LCDDIM || pm_cur_state == S_NORMAL);
}

static inline void restore_custom_brightness(void)
{
	if (pm_cur_state == S_LCDDIM &&
	    backlight_ops.get_custom_status())
		backlight_ops.custom_update();
}

static int power_execute(void *data)
{
	static const struct device_ops *ops = NULL;

	FIND_DEVICE_INT(ops, POWER_OPS_NAME);

	return ops->execute(data);
}

static void longkey_pressed()
{
	char *opt;
	unsigned int caps;

	_I("Power key long pressed!");
	cancel_lcdoff = 1;

	caps = display_get_caps(DISPLAY_ACTOR_POWER_KEY);

	if (display_has_caps(caps, DISPLAY_CAPA_LCDON)) {
		/* change state - LCD on */
		if (disp_plgn.pm_change_internal)
			disp_plgn.pm_change_internal(INTERNAL_LOCK_POWERKEY, LCD_NORMAL);
		(*pm_callback)(INPUT_POLL_EVENT, NULL);
	}

	if (!display_has_caps(caps, DISPLAY_CAPA_LCDOFF)) {
		_D("No poweroff capability!");
		return;
	}

	opt = POWEROFF_ACT;
	power_execute(opt);
}

static gboolean longkey_pressed_cb(void *data)
{
	longkey_pressed();
	longkey_timeout_id = 0;

	return G_SOURCE_REMOVE;
}

static unsigned long timediff_usec(struct timeval t1, struct timeval t2)
{
	unsigned long udiff;

	udiff = (t2.tv_sec - t1.tv_sec) * USEC_PER_SEC;
	udiff += (t2.tv_usec - t1.tv_usec);

	return udiff;
}

static inline void check_key_pair(int code, int new, int *old)
{
	if (new == *old)
		_E("key pair is not matched! (%d, %d)", code, new);
	else
		*old = new;
}

static inline void broadcast_lcdon_by_powerkey(void)
{
	dbus_handle_broadcast_dbus_signal(DEVICED_PATH_DISPLAY,
					DEVICED_INTERFACE_DISPLAY,
					SIGNAL_LCDON_BY_POWERKEY,
					NULL, NULL);
}

static inline void broadcast_lcdoff_by_powerkey(void)
{
	dbus_handle_broadcast_dbus_signal(DEVICED_PATH_DISPLAY,
					DEVICED_INTERFACE_DISPLAY,
					SIGNAL_LCDOFF_BY_POWERKEY,
					NULL, NULL);
}

static inline bool switch_on_lcd(void)
{
	if (current_state_in_on())
		return false;

	if (backlight_ops.get_lcd_power() == DPMS_ON) {
		if (ambient_get_state() == false)
			return false;
	}

	broadcast_lcdon_by_powerkey();

	lcd_on_direct(LCD_ON_BY_POWER_KEY);

	return true;
}

static inline void switch_off_lcd(void)
{
	if (!current_state_in_on())
		return;

	if (backlight_ops.get_lcd_power() == DPMS_OFF)
		return;

	broadcast_lcdoff_by_powerkey();

	lcd_off_procedure(LCD_OFF_BY_POWER_KEY);
}

static void check_key_combination(struct input_event *pinput)
{
	double press_time, diff_time;
	press_time = (pinput->time).tv_sec + USEC_TO_SEC((pinput->time).tv_usec);

	switch (key_combination) {
	case COMBINATION_SCREENCAPTURE:
		diff_time = press_time - combination_pressed_time;
		if (diff_time <= CAPTURE_COMBINATION_INTERVAL) {
			_I("Combination key : SCREENCAPTURE mode");
			skip_combination = true;
		}
		break;
	default:
		combination_pressed_time = press_time;
		return;
	}

}

static void start_key_combination(struct input_event *pinput)
{
	switch (pinput->code) {
	case KEY_POWER:
		key_combination |= KEY_COMBINATION_POWERKEY;
		break;
	case KEY_MENU:
		key_combination |= KEY_COMBINATION_MENUKEY;
		break;
	default:
		return;
	}

	check_key_combination(pinput);
}

static void stop_key_combination(struct input_event *pinput)
{
	if (pinput == NULL) {
		key_combination = KEY_COMBINATION_STOP;
		return;
	}

	switch (pinput->code) {
	case KEY_POWER:
		key_combination &= ~KEY_COMBINATION_POWERKEY;
		break;
	case KEY_MENU:
		key_combination &= ~KEY_COMBINATION_MENUKEY;
		break;
	default:
		_E("Thid code(%d) is not combination type", pinput->code);
		break;
	}
}

static void process_combination_key(struct input_event *pinput)
{
	if (pinput->value == KEY_PRESSED)
		start_key_combination(pinput);
	else if (pinput->value == KEY_RELEASED)
		stop_key_combination(pinput);
}

static int process_menu_key(struct input_event *pinput)
{
	int caps;

	caps = display_get_caps(DISPLAY_ACTOR_MENU_KEY);

	if (!display_has_caps(caps, DISPLAY_CAPA_LCDON)) {
		if (current_state_in_on())
			return false;
		_D("No lcd-on capability!");
		return true;
	} else if (pinput->value == KEY_PRESSED) {
		switch_on_lcd();
	}

	return false;
}

static bool release_short_powerkey(void)
{
// tv d - no, a - yes
	if (!display_conf.powerkey_doublepress && display_conf.lcd_always_on) {
		longkey_pressed();
		return false;
	}

	return true;
}

static int process_power_key(struct input_event *pinput)
{
	int ignore = true;
	static int value = KEY_RELEASED;
	unsigned int caps;

	caps = display_get_caps(DISPLAY_ACTOR_POWER_KEY);

	switch (pinput->value) {
	case KEY_RELEASED:
		check_key_pair(pinput->code, pinput->value, &value);

		ignore = release_short_powerkey();

		if (!display_has_caps(caps, DISPLAY_CAPA_LCDON))
			ignore = true;

		if (longkey_timeout_id > 0) {
			g_source_remove(longkey_timeout_id);
			longkey_timeout_id = 0;
		}

		break;
	case KEY_PRESSED:
		if (display_has_caps(caps, DISPLAY_CAPA_LCDON)) {
			skip_lcd_off = switch_on_lcd();
		} else {
			_D("No lcdon capability!");
			skip_lcd_off = false;
		}
		check_key_pair(pinput->code, pinput->value, &value);
		_I("power key pressed");
		pressed_time.tv_sec = (pinput->time).tv_sec;
		pressed_time.tv_usec = (pinput->time).tv_usec;
		if (key_combination == KEY_COMBINATION_POWERKEY) {
			/* add long key timer */
			longkey_timeout_id = g_timeout_add_seconds(
				    display_conf.longpress_interval,
				    longkey_pressed_cb, NULL);
		}
		if (skip_lcd_off)
			ignore = false;
		cancel_lcdoff = 0;

		break;
	case KEY_BEING_PRESSED:
		if (timediff_usec(pressed_time, pinput->time) >
		    (display_conf.longpress_interval * USEC_PER_SEC))
			longkey_pressed();
		break;
	}
	return ignore;
}

static int process_screenlock_key(struct input_event *pinput)
{
	if (pinput->value != KEY_RELEASED) {
		stop_key_combination(NULL);
		return true;
	}

	if (!current_state_in_on())
		return false;

	check_processes(S_NORMAL);
	check_processes(S_LCDDIM);

	if (!check_holdkey_block(S_NORMAL) && !check_holdkey_block(S_LCDDIM)) {
		delete_condition(S_NORMAL);
		delete_condition(S_LCDDIM);
		update_lcdoff_source(VCONFKEY_PM_LCDOFF_BY_POWERKEY);

		/* LCD off forcly */
		if (disp_plgn.pm_change_internal)
			disp_plgn.pm_change_internal(-1, LCD_OFF);
	}

	return true;
}

static void sound_vibrate_hardkey(void)
{
	/* device notify(vibrator) */
	/* sound(dbus) */
	/* Need to notify to deviced-vibrator. deviced-vibrator receives ChangedHardKey signal */
	dbus_handle_broadcast_dbus_signal(DEVICED_PATH_KEY,
					DEVICED_INTERFACE_KEY,
					SIGNAL_CHANGE_HARDKEY,
					NULL, NULL);
}

static void process_hardkey_backlight(struct input_event *pinput)
{
	int opt;

	_E("pinput->value : %d", pinput->value);
	if (pinput->value == KEY_PRESSED) {
		if (touch_pressed) {
			_I("Touch is pressed, then hard key is not working!");
			return;
		}
		/* Sound & Vibrate only in unlock state */
		if (__get_lock_screen_state() == VCONFKEY_IDLE_UNLOCK
		    || get_lock_screen_bg_state())
			sound_vibrate_hardkey();

		if (touchled && touchled->execute) {
			opt = TOUCHLED_PRESS;
			touchled->execute(&opt);
		}
	} else if (pinput->value == KEY_RELEASED) {
		/* if lockscreen is idle lock */
		if (__get_lock_screen_state() == VCONFKEY_IDLE_LOCK) {
			_D("Lock state, key backlight is off when phone is unlocked!");
			return;
		}

		if (touchled && touchled->execute) {
			opt = TOUCHLED_RELEASE;
			touchled->execute(&opt);
		}
	}
}

static int check_key(struct input_event *pinput, int fd)
{
	int ignore = true;

	process_combination_key(pinput);
	switch (pinput->code) {
	case KEY_MENU:
		ignore = process_menu_key(pinput);
		break;
	case KEY_POWER:
		ignore = process_power_key(pinput);
		if (current_state_in_on())
			ignore = false;
		break;
	case KEY_SCREENLOCK:
		ignore = process_screenlock_key(pinput);
		break;
	case KEY_BACK:
	case KEY_PHONE:
		stop_key_combination(NULL);
		if (current_state_in_on()) {
			process_hardkey_backlight(pinput);
			ignore = false;
		}
		break;
	case KEY_VOLUMEUP:
	case KEY_VOLUMEDOWN:
	case KEY_CAMERA:
	case KEY_EXIT:
	case KEY_CONFIG:
	case KEY_MEDIA:
	case KEY_MUTE:
	case KEY_PLAYPAUSE:
	case KEY_PLAYCD:
	case KEY_PAUSECD:
	case KEY_STOPCD:
	case KEY_NEXTSONG:
	case KEY_PREVIOUSSONG:
	case KEY_REWIND:
	case KEY_FASTFORWARD:
		stop_key_combination(NULL);
		if (current_state_in_on())
			ignore = false;
		break;
	case 0x1DB:
	case 0x1DC:
	case 0x1DD:
	case 0x1DE:
		stop_key_combination(NULL);
		break;
	default:
		stop_key_combination(NULL);
		ignore = false;
	}
#ifdef ENABLE_PM_LOG
	if (pinput->value == KEY_PRESSED)
		pm_history_save(PM_LOG_KEY_PRESS, pinput->code);
	else if (pinput->value == KEY_RELEASED)
		pm_history_save(PM_LOG_KEY_RELEASE, pinput->code);
#endif
	return ignore;
}

static int check_key_filter(void *data, int fd)
{
	struct input_event *pinput = data;
	int ignore = true;
	static int old_fd, code, value;

	assert(pinput);

	switch (pinput->type) {
	case EV_KEY:
		if (pinput->code == BTN_TOUCH &&
			pinput->value == KEY_RELEASED)
			touch_pressed = false;
		/*
		 * Normally, touch press/release events don't occur
		 * in lcd off state. But touch release events can occur
		 * in the state abnormally. Then touch events are ignored
		 * when lcd is off state.
		 */
		if (pinput->code == BTN_TOUCH && !current_state_in_on())
			break;
		if (pinput->code == code && pinput->value == value) {
			_E("Same key(%d, %d) is polled [%d,%d]",
				code, value, old_fd, fd);
		}
		old_fd = fd;
		code = pinput->code;
		value = pinput->value;

		ignore = check_key(pinput, fd);
		restore_custom_brightness();

		break;
	case EV_REL:
		ignore = false;
		break;
	case EV_ABS:
		if (current_state_in_on())
			ignore = false;

		if (ambient_get_condition() == true) {
			switch_on_lcd();
			ignore = false;
		}

		restore_custom_brightness();

		touch_pressed =
			(pinput->value == TOUCH_RELEASE ? false : true);
		break;
	case EV_SW:
		if (!get_glove_state || !switch_glove_key)
			break;
		if (pinput->code == SW_GLOVE &&
			get_glove_state() == GLOVE_MODE) {
			switch_glove_key(pinput->value);
		}
		break;
	}

	if (ignore)
		return 1;

	return 0;
}

/*
 * Default capability
 * powerkey := LCDON | LCDOFF | POWEROFF
 * homekey  := LCDON
 */
static struct display_actor_ops display_powerkey_actor = {
	.id	= DISPLAY_ACTOR_POWER_KEY,
	.caps	= DISPLAY_CAPA_LCDON |
		  DISPLAY_CAPA_LCDOFF |
		  DISPLAY_CAPA_POWEROFF,
};

static struct display_actor_ops display_menukey_actor = {
	.id	= DISPLAY_ACTOR_MENU_KEY,
	.caps	= DISPLAY_CAPA_LCDON,
};

static void keyfilter_init(void)
{
	display_add_actor(&display_powerkey_actor);
	display_add_actor(&display_menukey_actor);

	touchled = find_device(TOUCHLED_NAME);

	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
}

static void keyfilter_exit(void)
{
	unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
}

static void key_backlight_enable(bool enable)
{
	int opt;

	if (!touchled || !touchled->execute)
		return;

	if (enable)
		opt = TOUCHLED_DIRECT_ON;
	else
		opt = TOUCHLED_DIRECT_OFF;

	touchled->execute(&opt);
}

static const struct display_keyfilter_ops normal_keyfilter_ops = {
	.init			= keyfilter_init,
	.exit			= keyfilter_exit,
	.check			= check_key_filter,
	.set_powerkey_ignore	= NULL,
	.powerkey_lcdoff	= NULL,
	.backlight_enable	= key_backlight_enable,
};

const struct display_keyfilter_ops *keyfilter_ops = &normal_keyfilter_ops;

