/*
 * deviced
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>

#include "core/device-notifier.h"
#include "core/devices.h"
#include "display/core.h"
#include "display/util.h"
#include "display/display-ops.h"
#include <libsyscommon/dbus-system.h>

/* CoreApps */
#define COREAPPS_PATH_HOME                      "/Org/Tizen/Coreapps/home"
#define COREAPPS_INTERFACE_HOME                 "org.tizen.coreapps.home"

#define COREAPPS_PATH_SWIMMODE                  COREAPPS_PATH_HOME"/swimmode"
#define COREAPPS_INTERFACE_SWIMMODE             COREAPPS_INTERFACE_HOME".swimmode"
#define COREAPPS_SIGNAL_STATUS                  "status"

static const struct device_ops *touchscreen_ops;

static int swim_execute(void *data)
{
	int state, ret;
	static int old_state = -1;

	if (check_default(touchscreen_ops))
		return 0;

	state = (int) data;
	if (old_state == state)
		return 0;

	if (state) {
		_D("Stop touchscreen");
		ret = touchscreen_ops->stop(NORMAL_MODE);
	} else {
		_D("Start touchscreen");
		ret = touchscreen_ops->start(NORMAL_MODE);
	}

	old_state = state;
	return ret;
}

static void swimmode_signal_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	int val;
	pid_t pid;
	bool lcd_on = false;

	g_variant_get(param, "(i)", &val);
	pid = dbus_connection_get_sender_pid(conn, sender);
	_I("Request swim mode state %d by %d", val, pid);

	if (backlight_ops.get_lcd_power != DPMS_ON) {
		if (disp_plgn.pm_change_internal)
			disp_plgn.pm_change_internal(INTERNAL_LOCK_SWIM, S_NORMAL);
		lcd_on = true;
	}

	swim_execute((void *)val);

	if (lcd_on) {
		if (disp_plgn.pm_change_internal)
			disp_plgn.pm_change_internal(INTERNAL_LOCK_SWIM, S_LCDOFF);
	}
}

static int swim_restore_cb(void *data)
{
	_I("Restore swim mode");
	swim_execute((void *)0);

	return 0;
}

static void swim_init(void *data)
{
	int ret;

	touchscreen_ops = find_device("touchscreen");

	ret = subscribe_dbus_signal(NULL,
				COREAPPS_PATH_SWIMMODE,
				COREAPPS_INTERFACE_SWIMMODE,
				COREAPPS_SIGNAL_STATUS,
				swimmode_signal_handler,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);

	register_notifier(DEVICE_NOTIFIER_LONGKEY_RESTORE, swim_restore_cb);
}

static const struct device_ops swim_device_ops = {
	.name     = "swim",
	.init     = swim_init,
	.execute  = swim_execute,
};

DEVICE_OPS_REGISTER(&swim_device_ops)
