/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include "log.h"
#include "device-notifier.h"
#include "list.h"
#include "common.h"
#include "core/devices.h"
#include <libsyscommon/dbus-system.h>

struct device_notifier {
	bool deleted;
	enum device_notifier_type status;
	int (*func)(void *data);
};

static dd_list *device_notifier_list;
static guint idl;
static guint late_init_timer;

#define LATE_INIT_TIME		30
#define FIND_NOTIFIER(a, b, d, e, f) \
	DD_LIST_FOREACH(a, b, d) \
		if (e == d->e && f == (d->f))

int register_notifier(enum device_notifier_type status, int (*func)(void *data))
{
	dd_list *n;
	struct device_notifier *notifier;

	_I("notifier type=%d func=%p", status, func);

	if (!func) {
		_E("Invalid func address.");
		return -EINVAL;
	}

	FIND_NOTIFIER(device_notifier_list, n, notifier, status, func) {
		_E("Function is already registered. type=%d func=%p",
		    status, func);
		return -EINVAL;
	}

	notifier = calloc(1, sizeof(struct device_notifier));
	if (!notifier) {
		_E("Failed to malloc for notifier.");
		return -ENOMEM;
	}

	notifier->status = status;
	notifier->func = func;

	DD_LIST_APPEND(device_notifier_list, notifier);

	return 0;
}

int unregister_notifier(enum device_notifier_type status, int (*func)(void *data))
{
	dd_list *n;
	struct device_notifier *notifier;

	if (!func) {
		_E("Invalid func address.");
		return -EINVAL;
	}

	FIND_NOTIFIER(device_notifier_list, n, notifier, status, func) {
		_I("notifier type=%d func=%p", status, func);
		notifier->deleted = true;
	}

	return 0;
}

static gboolean delete_unused_notifier_cb(void *data)
{
	dd_list *n;
	dd_list *next;
	struct device_notifier *notifier;

	DD_LIST_FOREACH_SAFE(device_notifier_list, n, next, notifier) {
		if (notifier->deleted) {
			DD_LIST_REMOVE_LIST(device_notifier_list, n);
			free(notifier);
		}
	}

	idl = 0;
	return G_SOURCE_REMOVE;
}

void device_notify(enum device_notifier_type status, void *data)
{
	dd_list *n;
	struct device_notifier *notifier;

	DD_LIST_FOREACH(device_notifier_list, n, notifier) {
		if (!notifier->deleted && status == notifier->status) {
			if (notifier->func)
				notifier->func(data);
		}
	}

	if (!idl)
		idl = g_idle_add(delete_unused_notifier_cb, NULL);
}

void device_notify_once(enum device_notifier_type status, void *data)
{
	dd_list *n;
	struct device_notifier *notifier;

	DD_LIST_FOREACH(device_notifier_list, n, notifier) {
		if (!notifier->deleted && status == notifier->status) {
			if (notifier->func)
				notifier->func(data);

			notifier->deleted = true;
		}
	}

	if (!idl)
		idl = g_idle_add(delete_unused_notifier_cb, NULL);
}

static void late_init_stop(void)
{
	if (!late_init_timer)
		return;

	g_source_remove(late_init_timer);
	late_init_timer = 0;
}

static int booting_done(void *data)
{
	static int done;

	if (data == NULL)
		goto out;

	done = *(int *)data;
	if (!late_init_timer)
		return done;
	late_init_stop();

out:
	return done;
}

static gboolean late_init_timer_cb(void *data)
{
	int done;

	late_init_stop();
	done = booting_done(NULL);
	if (done)
		return G_SOURCE_REMOVE;
	_I("Late booting done.");
	done = TRUE;
	device_notify_once(DEVICE_NOTIFIER_BOOTING_DONE, (void *)&done);

	return G_SOURCE_REMOVE;
}

static void device_notifier_init(void *data)
{
	int ret;

	ret = check_systemd_active();
	if (ret == TRUE) {
		_I("Restart booting done.");
		return;
	}
	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);

	late_init_timer = g_timeout_add_seconds(LATE_INIT_TIME,
				late_init_timer_cb, NULL);
	if (!late_init_timer)
		_E("Failed to set 'late_init_timer'.");
}

static void device_notifier_exit(void *data)
{

}

static const struct device_ops notifier_device_ops = {
	.name     = "notifier",
	.init     = device_notifier_init,
	.exit     = device_notifier_exit,
};

DEVICE_OPS_REGISTER(&notifier_device_ops)

